-----------------------------------------------------------------------------
-- |
-- Module      :  Data.Relation.SetOfPairs
-- Copyright   :  (c) Joost Visser 2004
-- License     :  LGPL
-- 
-- Maintainer  :  joost.visser@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- An implementation of relations as sets of pairs.
--
-----------------------------------------------------------------------------

module Data.Relation.SetOfPairs (
  module Data.Relation.SetOfPairs,
  module Data.Set
) where

import Data.Set

-----------------------------------------------------------------------------

-- * Representation

-- | Type of relations
type Rel a b = Set (a,b)

instance Show a => Show (Set a) where
  show xs = "{" ++ show' (setToList xs) ++  "}"
            where
	      show' []     = ""
	      show' (x:xs) = show x ++ showl xs
	      showl []     = ""
              showl (x:xs) = "," ++ show x ++ showl xs

-----------------------------------------------------------------------------
-- * Building relations           

-- | Build an empty relation.              
emptyRel        :: Rel a b
emptyRel	        = emptySet

-- | Build a relation from a list of pairs.
mkRel           :: (Ord a, Ord b) => [(a,b)] -> Rel a b
mkRel pairs     = mkSet pairs

-- | Build identity relation, which contains an edge from each node to itself.
identityRel 		:: Ord a => Set a -> Rel a a
identityRel s	= mapSet (\x -> (x,x)) s

-- | Build total relation, which contains an edge from each node to 
--   each other node and to itself.
totalRel 		:: Ord a => Set a -> Rel a a
totalRel s		= mkSet [ (x,y) |  x <- l, y <- l ]
                          where
                            l = setToList s

-- | Build a chain relation of given number of numerals.
chainRel		:: (Enum n, Num n, Ord n) =>  n -> Rel n n
chainRel n 		= mkSet (map (\i -> (i,i+1)) [1..n])

-----------------------------------------------------------------------------
-- * Basic operations

-- | Obtain the domain of a relation
dom             :: Ord a => Rel a b -> Set a
dom xs          = mapSet fst xs

-- | Obtain the range of a relation
rng             :: Ord b => Rel a b -> Set b
rng xs          = mapSet snd xs

-- | Obtain all entities in a relation (union of domain and range)
ent             :: Ord a => Rel a a -> Set a
ent xs          = dom xs `union` rng xs

-- | Convert relation to a list of pairs.
pairs           :: (Ord a, Ord b) => Rel a b -> [(a,b)]
pairs r         = setToList r

-- | Take the inverse of a relation
inv             :: (Ord a, Ord b) => Rel a b -> Rel b a
inv xs 		= mapSet (\(x,y) -> (y,x)) xs

-- | Compose two relations
comp            :: (Ord a, Eq b, Ord c) => Rel b c -> Rel a b -> Rel a c
comp yz xy      = mkSet 
                    [ (x,z) | (x,y) <- setToList xy
                            , (y',z) <- setToList yz
			                , y==y'
                    ]

-----------------------------------------------------------------------------
-- * Closures

-- | Compute the reflective closure
reflClose       :: Ord a => Rel a a -> Rel a a
reflClose xs    = xs `union` mapSet (\x -> (x,x)) (ent xs)

-- | Compute the symmetric closure
symmClose       :: Ord a => Rel a a -> Rel a a
symmClose xs    = xs `union` mapSet (\(x,y) -> (y,x)) xs
		    
-- | Compute the transitive closure
transClose      :: Ord a => Rel a a -> Rel a a
transClose xs	= if xxs == xs then xs else (transClose xxs)
                  where 
		    xxs = union xs (comp xs xs)
		    
-- | Compute the reflexive, transitive closure
reflTransClose  :: Ord a => Rel a a -> Rel a a
reflTransClose xs
                = reflClose (transClose xs)

-- | Compute the reflexive, transitive, symmetric closure, i.e. the
--   induced equivalence relation.
equiv :: Ord b => Rel b b -> Rel b b
equiv r
  = reflTransClose (r `union` (inv r))


-----------------------------------------------------------------------------
-- * Selections (project, slice, chop)

-- | Projection of set through relation
project 	:: (Ord a, Ord b) => Set a -> Rel a b -> Rel a b
project	s r     = mkSet [ (x,y) | (x,y) <- setToList r, x `elementOf` s ]

-- | Projection of set backward through relation
projectBackward	:: (Ord a, Ord b) => Set b -> Rel a b -> Rel a b
projectBackward	s r
     		= mkSet [ (x,y) | (x,y) <- setToList r, y `elementOf` s ]

-- | Compute forward slice starting from seeds.
slice			:: Ord a => Set a -> Rel a a -> Rel a a
slice seed r		= sliceUntil seed emptySet r

-- | Compute forward slice starting from seeds, stopping at stoppers.
sliceUntil		:: Ord a => Set a -> Set a -> Rel a a -> Rel a a
sliceUntil seed stop r	= if (isEmptySet s) then emptyRel 
                                            else r' `union`
					           (sliceUntil seed' stop' r)
			  where
			    s = seed `minusSet` stop    -- live seeds
                	    r' = project s r            -- step
			    seed' = rng r'              -- new seeds
			    stop' = stop `union` s      -- extend stoppers

-- | Compute backward slice starting from seeds.
sliceBackward		:: Ord a => Set a -> Rel a a -> Rel a a
sliceBackward seed r	= inv (slice seed (inv r))

-- | Compute chop between seeds and sinks.
chop			:: Ord a => Set a -> Set a -> Rel a a -> Rel a a
chop seeds sinks r	= (slice seeds r)
			  `intersect`
		          (sliceBackward sinks r)

-----------------------------------------------------------------------------


