-----------------------------------------------------------------------------
-- |
-- Module      :  Data.Relation.SetOfPairsQuickCheck
-- Copyright   :  (c) Joost Visser 2004
-- License     :  LGPL
-- 
-- Maintainer  :  joost.visser@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Properties of the implementation of relations as sets of pairs.
--
-----------------------------------------------------------------------------

module Data.Relation.SetOfPairsQuickCheck where

import Data.Set
import Data.Relation.SetOfPairs
import Debug.QuickCheck

-----------------------------------------------------------------------------

instance (Ord a, Arbitrary a)
      => Arbitrary (Set a) where
  arbitrary = do xs <- arbitrary
                 return (mkSet xs)

-- | Inverting twice is identity.
prop_InvInv r	
  = classify (cardinality r == 0) "empty" $
    inv (inv r) == r
    where 
      types = r::Rel Int Integer

-- | Domain of inverse is range.
prop_DomInvRng :: Rel Int Integer -> Property
prop_DomInvRng r
  = classify (cardinality r == 0) "empty" $
    dom (inv r) == rng r    
    where 
      types = r::Rel Int Integer

-- | Range of inverse is domain.
prop_RngInvDom r
  = classify (cardinality r == 0) "empty" $
    rng (inv r) == dom r    
    where 
      types = r::Rel Int Integer

-- | Carrier of inverse is cariere of original.
prop_EntInvEnt r
  = classify (cardinality r == 0) "empty" $
    ent (inv r) == ent r    
    where 
      types = r::Rel Int Int

-- | Inverse distributes over composition.
prop_CompInvDist r1 r2
  = classify (cardinality r1 == 0 && cardinality r2 == 0) "both empty" $
    classify (cardinality r2 /= 0 && cardinality r2 /= 0) "none empty" $
    inv (r2 `comp` r1) == (inv r1) `comp` (inv r2)
    where 
      types = (r1::Rel Int Integer, r2::Rel Integer Double)

-- | 
close_prop close r 
  = classify (cardinality r == 0) "empty" $
    close r == close (close r)
    where
      types = r:: Rel Int Int
      
prop_ReflClose r
  = close_prop reflClose r
  
prop_SymmClose r
  = close_prop symmClose r
  
prop_TransClose r
  = close_prop transClose r

prop_ReflTransClose r
  = close_prop reflTransClose r

close_prop_Ent close r 
  = classify (cardinality r == 0) "empty" $
    ent r == ent (close r)
    where
      types = r:: Rel Int Int

prop_RelfCloseEnt r
  = close_prop_Ent reflClose r  
      
prop_SymmCloseEnt r
  = close_prop_Ent symmClose r      
  
prop_TransCloseEnt r
  = close_prop_Ent transClose r    
    
prop_ReflTransCloseEnt r
  = close_prop_Ent reflTransClose r      

prop_ReflCloseDomRng r
  = classify (cardinality r == 0) "empty" $
    dom (reflClose r) == rng (reflClose r)
    where
      types = r:: Rel Int Int  
      
prop_ReflCloseDomEnt r
  = classify (cardinality r == 0) "empty" $
    dom (reflClose r) == ent r
    where
      types = r:: Rel Int Int     

prop_ProjectInv s r
  = classify (cardinality s == 0 && cardinality r == 0) "both empty" $
    classify (cardinality s /= 0 && cardinality r /= 0) "none empty" $
    project s r == inv (projectBackward s (inv r))
    where
      types = (s::Set Int, r::Rel Int Integer)

prop_ChopInv seeds sinks r
  = classify (cardinality seeds /= 0 && cardinality sinks /= 0 && cardinality r /= 0)
        "none empty" $
    chop seeds sinks r == inv (chop sinks seeds (inv r))
    where
      types = (r::Rel Int Int)

prop_SliceIdentity s
  = classify (cardinality s /= 0) "non-empty" $
    slice s (identityRel s) == (identityRel s)
    where
      types = (s::Set Int)

prop_SliceTotal s
  = classify (cardinality s /= 0) "non-empty" $
    slice s (totalRel s) == (totalRel s)
    where
      types = (s::Set Int)

-----------------------------------------------------------------------------

